/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file msg.cpp
 *
 * Implements the cmsg class. For more information, see cmsg.h.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "receiveq.h"
#include "msg_factory.h"
#include "msg_data.h"
#include "msg.h"
#include "conn_mgr.h"
#include "id_table.h"
#include "debug.h"
#include "p4error.h"

#ifdef _WIN32
#include <winsock.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>     // for struct in_addr
#include <arpa/inet.h>      // for inet_ntoa
#endif

#ifdef __SVR4
#include <sys/byteorder.h>  // for ntohs, etc. on solaris
#endif

#include <cstdlib>

/*****************************************************************************
 * GLOBALS
 *****************************************************************************/
extern string glocal_user;

/*****************************************************************************
 * PUBLIC METHODS
 *****************************************************************************/

cmsg::cmsg(ushort plugin_id, const char *data, uint len) 
{
  incoming = false;
  byte_order = ehost;

  hdr.flags = 0;
  hdr.plugin_id = plugin_id;
  hdr.data_len = len;
  host_data_len = len;

  // TODO: we need a better random message id
  hdr.message_id = rand();

  // TODO: is this correct?
  // assert((!data && !len) || (data && len));
  this->data = 0;
  data_bytes_left = 0;

  assert(glocal_user.length() < kp4_username_len);
  strcpy(hdr.username, glocal_user.c_str());

  cdebug::instance() << acquire("msg_debug")
                     << "outgoing msg created " << *this
                     << release();
}

cmsg::~cmsg()
{
  // if data is 0, this is a noop
  if (host_data_len > 0)
    delete[] data;
}

ostream &operator << (ostream &ostr, const cmsg &msg)
{
  switch (msg.hdr.get_type()) {
  case kp4_hello:    ostr << "HELLO";    break;
  case kp4_ping:     ostr << "PING";     break;
  case kp4_ping_ack: ostr << "PING ACK"; break;
  case kp4_connect:  ostr << "CONNECT";  break;
  case kp4_discover: ostr << "DISCOVER"; break;
  case kp4_lka:      ostr << "LKA";      break;
  case kp4_lka_ack:  ostr << "LKA ACK";  break;
  case kp4_data:     ostr << "DATA";     break;
  default:           ostr << "(still constructing)"; break;
  }

  ostr << ", MID " << msg.hdr.message_id;
  ostr << ", from " << msg.hdr.username;

  ushort id = ((msg.byte_order == cmsg::ehost) ?
               msg.hdr.plugin_id : ntohs(msg.hdr.plugin_id));
  ostr << ", for plugin " << id;
  ostr << ", data len " << msg.host_data_len << ", ";

  ostr << ((msg.byte_order == cmsg::ehost) ? "host" : "network");
  ostr << " byte order";

  return ostr;
}


/*****************************************************************************
 * PROTECTED METHODS
 *****************************************************************************/
// lock's default value is true
const cconn &cmsg::send(const string &username, bool lock)
{
  cdebug::instance() << acquire("msg_debug")
                     << "msg::send() to " << username << " " << *this << " "
                     << username
                     << release();
  assert(!incoming);

  // are we sending to ourselves? if so, drop if system msg, enqueue if data
  if (username == glocal_user) {
    if (!hdr.is_type(kp4_data))
      return cconn::failed_instance();

    cmsg *copy = cmsg_factory::instance().new_msg(*this);

    // TODO: this has to be a static_cast to compile on windows (without
    // digging deep in vc++.net to turn on RTTI in compiling... but this should
    // REALLY REALLY REALLY be a dynamic cast! now i need to go wash my hands.
    cmsg_data *dataptr = reinterpret_cast<cmsg_data *>(copy);
    uint ret = creceiveq::instance().enqueue(dataptr);
    if (ret != kerr_success)
      delete copy;

    return cconn::failed_instance();  // TODO: shouldn't return failed
  }

  // send!
  return cconn_mgr::instance().send(*this, username, lock);
}


// lock's default value is true
uint cmsg::broadcast(bool lock)
{
  cdebug::instance() << acquire("msg_debug")
                     << "msg::broadcast() " << *this
                     << release();

  // add to id table so that we don't pass it on again
  if (!cid_table::instance().contains(hdr.message_id))
    cid_table::instance().add(hdr.message_id);

  hdr.set_flag(kp4_broadcast);
    return cconn_mgr::instance().broadcast(*this, lock);
}


bool cmsg::process()
{
  cdebug::instance() << acquire("msg_debug")
                     << "msg::process() " << *this
                     << release();

  assert(incoming);
  if (hdr.is_flag_set(kp4_broadcast)) {
    if (cid_table::instance().contains(hdr.message_id))
      return true;
    else
      cid_table::instance().add(hdr.message_id);
  }

  bool can_delete = virt_process();

  if (hdr.is_flag_set(kp4_broadcast)) {
    cmsg *copy = cmsg_factory::instance().new_msg(*this);
    // this is called from inside conn_mgr::receive, so don't acquire mutex
    copy->broadcast(false);

    delete copy;
  }

  return can_delete;
}


uint cmsg::recv_socket(int sock, uint &data_bytes_left)
{
  cdebug::instance() << acquire("msg_debug")
                     << "msg::recv_socket() on sock " << sock
                     << " host_data_len = " << host_data_len
                     << " this->data_bytes_left = " << this->data_bytes_left
                     << release();

  assert(incoming && sock >= 0);
  assert(host_data_len >= this->data_bytes_left);

  // receive data into data buffer
  if (this->data_bytes_left > 0) {
    int read = recv(sock, data + host_data_len - this->data_bytes_left,
                    this->data_bytes_left, 0);
    if (read == -1 || read == 0)
      return kerr_couldnt_receive;
    cdebug::instance() << acquire("msg_debug")
                       << "read " << read << " bytes"
                       << release();
    this->data_bytes_left -= read;
  }

  assert(this->data_bytes_left >= 0);
  data_bytes_left = this->data_bytes_left;
  return kerr_success;
}


void cmsg::ntoh()
{
  assert(byte_order == enetwork);

#if (defined(_WIN32) || defined(__SVR4))
  convert_byteorder(cmsg::my_ntohs, cmsg::my_ntohl);
#else
  convert_byteorder(ntohs, ntohl);
#endif  // _WIN32 || __SVR4

  byte_order = ehost;
}

void cmsg::hton()
{
  assert(byte_order == ehost);

#if (defined(_WIN32) || defined(__SVR4))
  convert_byteorder(cmsg::my_htons, cmsg::my_htonl);
#else
  convert_byteorder(htons, htonl);
#endif  // _WIN32 || __SVR4

  byte_order = enetwork;
}


void cmsg::convert_byteorder(cmsg::short_fn sfn, cmsg::long_fn lfn)
{
  hdr.plugin_id = sfn(hdr.plugin_id);
  hdr.data_len = lfn(hdr.data_len);
 
  // TODO: convert message_id byteorder (but how, it's a double!)
}


cmsg::cmsg(const p4_header &hdr) : hdr(hdr)
{
  incoming = true;
  byte_order = enetwork;

  assert(hdr.flags != 0);

  // incoming message so we know it's in network byte order
  host_data_len = ntohl(hdr.data_len);

  if (host_data_len > 0) {  // do we need to allocate room for the data?
    data = new char[host_data_len];
    assert(data);
    data_bytes_left = host_data_len;

  } else {
    data = 0;
    data_bytes_left = 0;
  }

  cdebug::instance() << acquire("msg_debug")
                     << "incoming msg created " << *this
                     << release();
}


// copy ctor (protected!)
cmsg::cmsg(const cmsg &msg) : hdr(msg.hdr)
{
  incoming = msg.incoming;
  byte_order = msg.byte_order;
  data_bytes_left = msg.data_bytes_left;
  host_data_len = msg.host_data_len;

  if (host_data_len > 0) {
    data = new char[host_data_len];
    memcpy(data, msg.data, host_data_len);

  } else {
    data = 0;
  }
  cdebug::instance() << acquire("msg_debug")
                     << "msg copied" << *this
                     << release();
}
